// Mike Woodring
// http://staff.develop.com/woodring
//
// This variation of the tpcontrol program provides a UI that
// allows you to use BeginInvoke to 'kidnap' a worker thread,
// or perform an asyncrhonous I/O operation to 'kidnap' an
// I/O thread.
//
// On the V1 version of the platform, this program demonstrates
// that the max worker thread count and max I/O thread count
// are maintained independent of one another, as opposed to
// a semantic of "max worker threads = X, and of those, Y can
// be devoted to I/O operations".
//
using System;
using System.IO;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Threading;

class IoState
{
    public IoState( NetworkStream stream, byte[] buf )
    {
        this.stream = stream;
        this.buf = buf;
    }

    public readonly NetworkStream stream;
    public readonly byte[] buf;
}

delegate void SuspendProc();

class IoTest
{
    static ICorThreadPool  clrThreadPool = CLRThreadPool.Controller;
    static ManualResetEvent evtExitServer = new ManualResetEvent(false);
    static AutoResetEvent evtSendData = new AutoResetEvent(false);

    static void Main(string[] args)
    {
        ShowThreadPoolStats("initially");
        // AdjustThreadPool();
        // ShowThreadPoolStats("adjusted");

        // Start a thread listening on a socket in this app
        // so that we have something to perform blocking I/O
        // against that we know will take a long time to
        // return data from.
        //
        Thread t = StartServer();

        // Connect to the server.
        //
        TcpClient c = new TcpClient("localhost", 8080);
        NetworkStream strmToServer = c.GetStream();

        AsyncCallback cb = new AsyncCallback(OnReadComplete);

        for( int n = 0; n < 100; n++ )
        {
            byte[] buf = new byte[30];
            strmToServer.BeginRead( buf, 0, buf.Length,
                                    cb, new IoState(strmToServer, buf) );
        }

        Console.WriteLine("Press key to [t]ransmit, [q]uit, [k]idnap worker, other to show stats");

        bool done = false;

        while( !done)
        {
            string cmd = Console.ReadLine();

            switch( cmd )
            {
                case "t":
                case "T":
                    evtSendData.Set();
                    Thread.Sleep(750); // Allow it to happen.
                    ShowThreadPoolStats("after blocking I/O");
                break;

                case "k":
                case "K":
                    SuspendProc threadSuspend = new SuspendProc(SuspendCaller);
                    threadSuspend.BeginInvoke(null, null);
                    Thread.Sleep(750); // Allow it to happen.
                    ShowThreadPoolStats("after blocking BeginInvoke");
                break;

                case "q":
                case "Q":
                    done = true;
                    evtExitServer.Set();
                    t.Join();
                break;

                default:
                    ShowThreadPoolStats("current");
                break;
            }
        }
    }

    static void SuspendCaller()
    {
        Console.WriteLine("[{0}] Suspending calling thread", Thread.CurrentThread.GetHashCode());
        Thread.CurrentThread.Suspend();
    }

    static void ShowThreadPoolStats( string msg )
    {
        uint maxWorkerThreads;
        uint availWorkerThreads;
        uint maxIOThreads;
        uint availIOThreads;

        clrThreadPool.GetMaxThreads(out maxWorkerThreads, out maxIOThreads);
        clrThreadPool.GetAvailableThreads(out availWorkerThreads, out availIOThreads);

        Console.WriteLine("Thread pool statistics: {0}", msg);
        Console.WriteLine("  Max worker threads = {0}, available = {1}", maxWorkerThreads, availWorkerThreads);
        Console.WriteLine("  Max I/O threads = {0}, available = {1}", maxIOThreads, availIOThreads);
    }

    static void AdjustThreadPool()
    {
        uint maxWorkerThreads;
        uint maxIOThreads;

        clrThreadPool.GetMaxThreads(out maxWorkerThreads, out maxIOThreads);

        maxWorkerThreads += (maxWorkerThreads / 4);
        maxIOThreads += (maxIOThreads / 4);

        Console.WriteLine( "Setting to max worker threads to {0}, max IO threads to {1}",
                            maxWorkerThreads, maxIOThreads );

        clrThreadPool.SetMaxThreads(maxWorkerThreads, maxIOThreads);
    }

    static void OnReadComplete( IAsyncResult ar )
    {
        IoState state = (IoState)ar.AsyncState;
        int cbRead = state.stream.EndRead(ar);
        int tid = Thread.CurrentThread.GetHashCode();

        Console.WriteLine( "[{0}] Finished reading {1} bytes.",
                           tid, cbRead );

        if( cbRead > 0 )
        {
            string data = Encoding.ASCII.GetString(state.buf, 0, cbRead);
            Console.Write("[{0}] Data: {1}", tid, data);
        }

        Thread.Sleep(Timeout.Infinite);
    }

    static Thread StartServer()
    {
        Thread t = new Thread(new ThreadStart(ServerProc));
        t.IsBackground = true;
        t.Start();
        Thread.Sleep(100); // Hack: give the new thread chance to start.
        return(t);
    }

    static void ServerProc()
    {
        // Start listening on port 8080.
        //
        TcpListener l = new TcpListener(8080);
        l.Start();

        // Wait for a client to connect.
        //
        Socket sock = l.AcceptSocket();

        // Setup a NetworkStream on the socket,
        // and a StreamWriter on the NetworkStream.
        //
        NetworkStream strm = new NetworkStream(sock, false);
        StreamWriter w = new StreamWriter(strm);
        int tid = Thread.CurrentThread.GetHashCode();

        bool done = false;

        for( int n = 0; !done; n++ )
        {
            WaitHandle[] wh = {evtExitServer, evtSendData};

            switch( WaitHandle.WaitAny(wh) )
            {
                case 0:
                    done = true;
                break;

                case 1:
                    w.WriteLine("tid {0} #{1}", tid, n);
                    w.Flush();
                break;
            }
        }

        w.Close();
        strm.Close();
        sock.Close();
    }
}
