// Mike Woodring
// http://staff.develop.com/woodring
//
// This sample program demonstrates how to adjust the CLR-managed
// thread pool at runtime.
//
// To play with this program, build and run it.  Then bring up
// the Windows task list.  Use View/Select Columns to view the
// Thread Count for each process.  When tpcontrol is started,
// it will display the current thread pool stats, set
// the thread pool limits to 2x the current limit, redisplay
// the pool stats, and then wait for you to press ENTER to start
// the test.  Note the thread count at this time to see what the
// initial thread count it.
//
// When you press ENTER again, this program will just queue up
// 100 requests to the thread pool for a callback.  The callback
// function blocks the caller forever, causing the thread pool
// manager to add new threads to the pool to meet demand until the
// configured maximums have been reached.  Once you start the
// test, press ENTER at any time to take a snapshot of the pool's
// current stats (in particular, the available thread count).  Press
// 'q' followed by ENTER to exit the program.
//
// Note, there's a block of code in Main that's conditionally compiled
// out using #if (false).  Set that to true to cause the pool to be
// increased every third time you show the current pool stats.  This just
// proves that the pool can be adjusted over the life of a program,
// and not just strictly at startup.
//
// Because this program uses COM interop to reach out and call
// MSCOREE, the tpcontrol.exe assembly needs sufficient permissions
// to operate correctly.  Running tpcontrol.exe from your local file
// system is the easiest way to have sufficient permissions.
//
// To do its job, this program takes advantage of an interface
// called ICorThreadpool that is implemented by the runtime.  Because
// this interface is mentioned in mscoree.h, but not documented in
// mscoree.tlb, an explicit interop shim is used.  See ICorThreadPool.cs
// in this project for details.
//
// There is a sub-project called iotest.exe that demonstrates that the
// "I/O" threads and "worker" threads managed by the thread pool are
// accounted for separately.  In other words, if ICorThreadPool.GetMaxThreads
// returns 25 and 25, then that's a total of 50 threads (as opposed to
// saying there are 25 threads max, of which up to 25 can be devoted to
// I/O).
//
using System;
using System.Threading;
using System.Runtime.InteropServices;

class ThreadPoolControlApp
{
    private static ICorThreadPool clrThreadPool = CLRThreadPool.Controller;

    static void Main()
    {
        ShowThreadPoolStats("initially");  // Show default stats.
        AdjustThreadPool();
        ShowThreadPoolStats("after adjusting");  // Show new stats.

        Console.WriteLine("Check thread count in taskmgr, then press ENTER to start test");
        Console.ReadLine();
        TestPool();
        
        Console.WriteLine("Press q to exit, ENTER to show stats");

        int n = -1;
        n++; // Sets n to zero and avoids the 'var not used' warning
             // when the #if block of code below isn't being compiled.

        while( Console.ReadLine() != "q" )
        {
            ShowThreadPoolStats("current");

#if (false)
            // Every 3rd time, adjust the pool again.
            //
            n++;

            if( (n % 3) == 0 )
            {
                AdjustThreadPool();
            }
#endif
        }
    }

    static void ShowThreadPoolStats( string msg )
    {
        uint maxWorkerThreads;
        uint availWorkerThreads;
        uint maxIOThreads;
        uint availIOThreads;

        clrThreadPool.GetMaxThreads(out maxWorkerThreads, out maxIOThreads);
        clrThreadPool.GetAvailableThreads(out availWorkerThreads, out availIOThreads);

        Console.WriteLine("Thread pool statistics: {0}", msg);
        Console.WriteLine("  Max worker threads = {0}, available = {1}", maxWorkerThreads, availWorkerThreads);
        Console.WriteLine("  Max I/O threads = {0}, available = {1}", maxIOThreads, availIOThreads);
    }

    static void AdjustThreadPool()
    {
        uint maxWorkerThreads;
        uint maxIOThreads;

        clrThreadPool.GetMaxThreads(out maxWorkerThreads, out maxIOThreads);

        maxWorkerThreads *= 2;
        maxIOThreads *= 2;

        Console.WriteLine( "Setting to max worker threads to {0}, max IO threads to {1}",
                            maxWorkerThreads, maxIOThreads );

        clrThreadPool.SetMaxThreads(maxWorkerThreads, maxIOThreads);
    }

    static void TestPool()
    {
        WaitCallback cb = new WaitCallback(ThreadProc);

        // Queue up 100 requests to the thread pool.
        //
        for( int n = 0; n < 100; n++ )
        {
            ThreadPool.QueueUserWorkItem(cb, null);
        }
    }

    static void ThreadProc( object state )
    {
        Thread.Sleep(Timeout.Infinite);
    }
}
